@@ -18,6 +18,7 @@ use std::io;
 use std::io::prelude::*;
 use std::io::BufReader;
 use std::path::Path;
+use std::fmt::Debug;
 use std::fs::File;
 use std::str;
 
@@ -34,6 +35,7 @@ use openpgp::cert::prelude::*;
 use openpgp::parse::{
     Parse,
     PacketParserResult, PacketParser,
+    buffered_reader::BufferedReader,
 };
 use openpgp::serialize::Serialize;
 use openpgp::serialize::stream::{
@@ -358,6 +360,16 @@ impl AutocryptHeaders {
     pub fn from_reader<R: io::Read>(reader: R) -> Result<Self> {
         Self::from_lines(BufReader::new(reader).lines())
     }
+
+    /// Parses an autocrypt header.
+    ///
+    /// `reader` should be all of a mail's headers.
+    pub fn from_buffered_reader<R, C>(mut reader: R) -> Result<Self>
+    where R: BufferedReader<C>,
+          C: Debug + Send + Sync
+    {
+        Self::from_bytes(reader.data_eof()?)
+    }
 }
 
 /// Holds an Autocrypt Setup Message.
@@ -453,7 +465,7 @@ impl AutocryptSetupMessage {
 
 
     // Generates a new passcode in "numeric9x4" format.
-    fn passcode_gen() -> Password {
+    fn passcode_gen() -> Result<Password> {
         use openpgp::crypto::mem;
         // Generate a random passcode.
 
@@ -461,7 +473,7 @@ impl AutocryptSetupMessage {
         // approximately 119 bits of information.  120 bits = 15
         // bytes.
         let mut p_as_vec = mem::Protected::from(vec![0; 15]);
-        openpgp::crypto::random(&mut p_as_vec[..]);
+        openpgp::crypto::random(&mut p_as_vec[..])?;
 
         // Turn it into a 128-bit number.
         let mut p_as_u128 = 0u128;
@@ -480,21 +492,22 @@ impl AutocryptSetupMessage {
             p_as_u128 /= 10;
         }
 
-        p.into()
+        Ok(p.into())
     }
 
     /// If there is no passcode, generates one.
-    fn passcode_ensure(&mut self) {
+    fn passcode_ensure(&mut self) -> Result<()> {
         if self.passcode.is_some() {
-            return;
+            return Ok(());
         }
 
-        let passcode = Self::passcode_gen();
+        let passcode = Self::passcode_gen()?;
         self.passcode_format = Some("numeric9x4".into());
         self.passcode_begin = passcode.map(|p| {
             Some(str::from_utf8(&p[..2]).unwrap().into())
         });
         self.passcode = Some(passcode);
+        Ok(())
     }
 
     /// Generates the Autocrypt Setup Message.
@@ -508,7 +521,7 @@ impl AutocryptSetupMessage {
         // SEIP packet contains a literal data packet whose content is
         // the inner message.
 
-        self.passcode_ensure();
+        self.passcode_ensure()?;
 
         let mut headers : Vec<(&str, &str)> = Vec::new();
         if let Some(ref format) = self.passcode_format {
@@ -786,6 +799,7 @@ impl<'a> AutocryptSetupMessageParser<'a> {
         // Get the MDC packet.
         if let PacketParserResult::Some(pp) = ppr {
             match pp.packet {
+                #[allow(deprecated)]
                 Packet::MDC(_) => (),
                 ref p => return
                     Err(Error::MalformedMessage(
@@ -826,6 +840,7 @@ impl<'a> AutocryptSetupMessageParser<'a> {
 
 #[cfg(test)]
 mod test {
+    use sequoia_openpgp::parse::buffered_reader;
     use super::*;
 
     use openpgp::policy::StandardPolicy as P;
@@ -852,7 +867,7 @@ mod test {
             .expect("Failed to parse key material.");
         assert_eq!(cert.fingerprint(),
                    "156962B0F3115069ACA970C68E3B03A279B772D6".parse().unwrap());
-        assert_eq!(cert.userids().next().unwrap().value(),
+        assert_eq!(cert.userids().next().unwrap().userid().value(),
                    &b"holger krekel <holger@merlinux.eu>"[..]);
 
 
@@ -877,7 +892,7 @@ mod test {
             .expect("Failed to parse key material.");
         assert_eq!(cert.fingerprint(),
                    "D4AB192964F76A7F8F8A9B357BD18320DEADFA11".parse().unwrap());
-        assert_eq!(cert.userids().next().unwrap().value(),
+        assert_eq!(cert.userids().next().unwrap().userid().value(),
                    &b"Vincent Breitmoser <look@my.amazin.horse>"[..]);
 
 
@@ -902,7 +917,7 @@ mod test {
             .expect("Failed to parse key material.");
         assert_eq!(cert.fingerprint(),
                    "4F9F89F5505AC1D1A260631CDB1187B9DD5F693B".parse().unwrap());
-        assert_eq!(cert.userids().next().unwrap().value(),
+        assert_eq!(cert.userids().next().unwrap().userid().value(),
                    &b"Patrick Brunschwig <patrick@enigmail.net>"[..]);
 
         let ac2 = AutocryptHeaders::from_bytes(
@@ -912,6 +927,40 @@ mod test {
         assert_eq!(ac, ac2);
     }
 
+    #[test]
+    fn parse_from_buffered_reader() {
+        // This test relies on 'from_bytes(...)' parser which, in case of a bug,
+        // may cause this test to fail as well.
+        let ac1_bytes = AutocryptHeaders::from_bytes(
+            &include_bytes!("../tests/data/hpk.txt")[..]
+        )
+        .unwrap();
+        let ac2_bytes = AutocryptHeaders::from_bytes(
+            &include_bytes!("../tests/data/vincent.txt")[..]
+        )
+        .unwrap();
+        let ac3_bytes = AutocryptHeaders::from_bytes(
+            &include_bytes!("../tests/data/patrick_unfolded.txt")[..]
+        )
+        .unwrap();
+
+        let ac1_br = AutocryptHeaders::from_buffered_reader(
+            buffered_reader::File::open("tests/data/hpk.txt").unwrap()
+        )
+        .unwrap();
+        let ac2_br = AutocryptHeaders::from_buffered_reader(
+            buffered_reader::File::open("tests/data/vincent.txt").unwrap()
+        )
+        .unwrap();
+        let ac3_br = AutocryptHeaders::from_buffered_reader(
+            buffered_reader::File::open("tests/data/patrick_unfolded.txt").unwrap()
+        )
+        .unwrap();
+
+        assert_eq!(ac1_bytes, ac1_br);
+        assert_eq!(ac2_bytes, ac2_br);
+        assert_eq!(ac3_bytes, ac3_br);
+    }
 
     #[test]
     fn decode_gossip() {
@@ -935,7 +984,7 @@ mod test {
             .expect("Failed to parse key material.");
         assert_eq!(cert.fingerprint(),
                    "C4BC2DDB38CCE96485EBE9C2F20691179038E5C6".parse().unwrap());
-        assert_eq!(cert.userids().next().unwrap().value(),
+        assert_eq!(cert.userids().next().unwrap().userid().value(),
                 &b"Daniel Kahn Gillmor <dkg@fifthhorseman.net>"[..]);
 
     }
@@ -952,7 +1001,7 @@ mod test {
         let passcode_len = 36 + (36 / 4 - 1);
 
         for _ in 0..samples {
-            let p = AutocryptSetupMessage::passcode_gen();
+            let p = AutocryptSetupMessage::passcode_gen().unwrap();
             p.map(|p| {
                 assert_eq!(p.len(), passcode_len);
 
